/*
 * Created on Jan 13, 2004
 */
package edu.umd.cs.daveho.eclipse.autocvs;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.ResourceBundle;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IResourceDeltaVisitor;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPluginDescriptor;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Plugin;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.team.internal.ccvs.core.CVSException; // XXX: internal API
import org.eclipse.team.internal.ccvs.core.ICVSFile;
import org.eclipse.team.internal.ccvs.core.ICVSResource; // XXX: internal API
import org.eclipse.team.internal.ccvs.core.resources.CVSWorkspaceRoot; // XXX: internal API
import org.eclipse.team.internal.ccvs.core.syncinfo.ResourceSyncInfo;
import org.eclipse.ui.IStartup;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.plugin.AbstractUIPlugin;

/**
 * The AutoCVSPlugin runs CVS commands automatically to keep
 * a project's contents up to date with the CVS repository.
 * 
 * @author David Hovemeyer
 */
public class AutoCVSPlugin extends AbstractUIPlugin implements IStartup {
	// Constants
	/** Nature id for the AutoCVS nature. */
	public static final String AUTO_CVS_NATURE = "edu.umd.cs.daveho.eclipse.autocvs.autoCVSNature";
	/** Nature id for the AutoSync nature. */
	public static final String AUTO_SYNC_NATURE = "edu.umd.cs.daveho.eclipse.autocvs.autoSyncNature";
	
	// Fields
	private static AutoCVSPlugin plugin;
	private ResourceBundle messageBundle;
	private IResourceChangeListener listener;
	private EventLog eventLog;
	
	/**
	 * Set to true if any automatic CVS operation fails.
	 * Upon the failure of an AutoCVS operation,
	 * we inform the user and disable automatic execution
	 * for the remainder of the Eclipse session. 
	 */
	private boolean failedOperation;
	
	/**
	 * Object used to accumulate modifications made to a project's files.
	 */
	private static class ModificationBundle {
		private ArrayList addedFiles = new ArrayList();
		private ArrayList modifiedFiles = new ArrayList();
		
		public void addedFile(IResource resource) {
			addedFiles.add(resource);
			modifiedFiles.add(resource); // added resources also need to be committed
		}
		
		public void modifiedFile(IResource resource) {
			modifiedFiles.add(resource);
		}
		
		public boolean hasAddedFiles() { return !addedFiles.isEmpty(); }
		public boolean hasModifiedFiles() { return !modifiedFiles.isEmpty(); }
		
		public IResource[] getFilesToAdd() {
			return (IResource[]) addedFiles.toArray(new IResource[0]);
		}
		
		public IResource[] getFilesToCommit() {
			return (IResource[]) modifiedFiles.toArray(new IResource[0]);
		}
	}
	
	/**
	 * A resource change listener to capture resource modification events.
	 * 
	 * @author David Hovemeyer
	 */
	private class AutoCVSResourceModificationListener implements IResourceChangeListener {
		public void resourceChanged(IResourceChangeEvent event) {
			
			// Nothing to do if an automatic CVS operation has failed earlier.
			if (hasFailedOperation())
				return;
			
			int type = event.getType();
			Debug.print("Resource change event type = " + type);
			
			// The resource delta represents all changed resources
			IResourceDelta delta = event.getDelta();
			if (delta == null)
				return; // No resources modified - could be project being deleted
			
			// Accumulate resources that have been added or had their contents changed
			final ModificationBundle bundle = new ModificationBundle();
			IResourceDeltaVisitor visitor = new IResourceDeltaVisitor(){
				public boolean visit(IResourceDelta delta) throws CoreException {
					int kind = delta.getKind();
					int flags = delta.getFlags();
					IResource resource = delta.getResource();
					IProject project = resource.getProject();
					
					try {
						if (project == null)
							return true; // This is the workspace root. Continue visiting children.
						
						if (!project.isOpen())
							return false; // Project is not open yet - don't continue visiting children
						
						// Do not proceed if AutoSync is disabled.
						boolean autoSync = AutoCVSPlugin.getPlugin().getAutoSync(project);
						if (!autoSync)
							return false; // No need to continue visiting children
						
						if (project == resource && (kind & IResourceDelta.ADDED) != 0) {
							// Project is being added.
							// We see this when a project is checked out for the first time.
							// In this case, we definitely do not want to perform
							// any CVS operations on it.
							Debug.print("Project " + project.getName() + " added");
							return false; // Do not continue visiting children
						}
						
						// There are several things to do here.
						// - If the resource's project is not AutoCVS managed,
						//		ignore it.
						// - If the resource is of an ignored resource type,
						//		ignore it.
						// - If the resource has been added:
						//		Schedule it to be added to CVS.
						// - If the resource's content is changed:
						//		1. If it isn't in CVS, schedule it to be added (a safeguard: should never happen)
						//		2. Schedule it for a commit
						
						if (!hasAutoCVSNature(project))
							return false; // Not an AutoCVS project: no need to continue visiting children

						if (isCVSIgnored(resource))
							return true; // Resource ignored by CVS
						
						// Do not add or commit .class files under any circumstances
						if (!(resource instanceof IContainer) && resource.getName().endsWith((".class"))) {
							Debug.print("Class file "+resource.getName() + " is not ignored by CVS");
							return false;
						}
						
						if ((kind & IResourceDelta.ADDED) != 0){
							bundle.addedFile(resource);
						} else if ((flags & IResourceDelta.CONTENT) != 0) {
							if (!isCVSManaged(resource)) {
								// Paranoid case: resource is unexpectedly unmanaged
								AutoCVSPlugin.getPlugin().getEventLog().logMessage(
										"Warning: detected change to non-CVS-managed resource " +
										resource.getName() +
										": scheduling it for addition");
								bundle.addedFile(resource);
								// Note: added resources are automatically scheduled for commit
							} else {
								// Usual case: changed resource is managed.
								// Schedule it for commit.
								bundle.modifiedFile(resource);
							}
						}
						
					} catch (Throwable e) {
						eventLog.logError("Exception checking resource delta", e);
						// TODO: use marker to notify user?
					}
					return true;
				}
			};
			
			try {
				delta.accept(visitor);
				
				if(!bundle.hasAddedFiles() && !bundle.hasModifiedFiles()) {
					Debug.print("No resources added, no managed resources were changed.");
					return;
				}
				
				// Added resources must also be committed. (?)
				//changedResources.addAll(addedResources);

				// Create an OperationContext object to record the
				// success or failure of the add and commit operations.
				// This will also have the effect of circumventing the
				// commit operation if the add fails.
				CVSOperations.SimpleContext context = new CVSOperations.SimpleContext() {
					public void failure(Shell parent) {
						setFailedOperation(true);
					}
					public String getConsequenceMessage(int statusCode) {
						if (statusCode == CVSOperations.FAILED_UP_TO_DATE_CHECK)
							return getMessage("failed.commit.notuptodate");
						else if (statusCode == CVSOperations.UNRESOLVED_MERGE_CONFLICT)
							return getMessage("failed.commit.mergeconflict");
						else
							return null;
					}
				};
				
				// Add added resources.
				CVSOperations.add(bundle.getFilesToAdd(), CVSOperations.ASYNC, context);
				
				// Commit changed resources.
				CVSOperations.commit(bundle.getFilesToCommit(), CVSOperations.ASYNC, context);
			} catch (Exception e) {
				eventLog.logError("Exception in automatic commit", e);
				// TODO: use marker to notify user?
				return;
			}
		}
	}
	
	/**
	 * Constructor.
	 * @param descriptor the plugin descriptor
	 */
	public AutoCVSPlugin(IPluginDescriptor descriptor) {
		super(descriptor);

		Debug.print("AutoCVSPlugin: construction");
		
		// Save the plugin for later access
		plugin = this;
	}

	public void startup() throws CoreException {
		Debug.print("AutoCVS regular startup");
		
		// Create event log.
		this.eventLog = new EventLog();
		
		// Load the resource bundle with the messages
		this.messageBundle = ResourceBundle.getBundle("edu.umd.cs.daveho.eclipse.autocvs.Messages");
		
		// Install our resource modification listener
		this.listener = new AutoCVSResourceModificationListener();
		ResourcesPlugin.getWorkspace().addResourceChangeListener(listener);
		
		// Force the CVS core plugin to load
		Plugin cvsPlugin = Platform.getPlugin("org.eclipse.team.cvs.core");
		if (cvsPlugin != null){
			Debug.print("Successfully got handle of CVS plugin");
		}

		// Update projects in workspace
		try {
			autoUpdateWorkspaceProjects();
		} catch (CoreException e) {
			eventLog.logError("Exception updating workspace projects", e);
		} catch (InvocationTargetException e) {
			eventLog.logError("Exception updating workspace projects", e);
		} catch (InterruptedException e) {
			eventLog.logError("Exception updating workspace projects", e);
		}
	}

	public void earlyStartup() {
		Debug.print("AutoCVS Plugin early startup");
	}
	
	/**
	 * Get a message from the message bundle.
	 * 
	 * @param key the key of the message to retrieve
	 * @return the message
	 */
	public static String getMessage(String key) {
		return getPlugin().messageBundle.getString(key);
	}
	
	/**
	 * Determine whether given resource is managed by CVS.
	 * @param resource the resource
	 * @return true if resource is managed by CVS, false otherwise
	 * @throws CVSException
	 */
	static boolean isCVSManaged(IResource resource) throws CVSException {
		boolean result;
		ICVSResource cvsResource = CVSWorkspaceRoot.getCVSResourceFor(resource);
		if (cvsResource == null) {
			result =false;
		} else {
			result = cvsResource.isManaged();
		}
		return result;
	}

	/**
	 * Determine whether given resource is/should be ignored by CVS.
	 * @param resource the resource
	 * @return true if resource should be ignored, false if not
	 * @throws CVSException
	 */
	static boolean isCVSIgnored(IResource resource) throws CVSException {
		ICVSResource mResource = CVSWorkspaceRoot.getCVSResourceFor(resource);
		return mResource != null && mResource.isIgnored();
	}
	
	/**
	 * Determine whether given resource has an unresolved merge conflict.
	 * @param resource the resource
	 * @return true if the file has a merge conflict, false if not
	 * @throws CVSException
	 */
	static boolean hasMergeConflict(IResource resource) throws CVSException {
		if (!(resource instanceof IFile))
			return false;
		IFile file = (IFile) resource;
		ICVSFile cvsFile = CVSWorkspaceRoot.getCVSFileFor(file);
		if (cvsFile == null)
			return false;
		ResourceSyncInfo info = cvsFile.getSyncInfo();
		return info != null && info.isMergedWithConflicts();
	}
	
	/**
	 * Attempt an automatic CVS update of given project
	 * @param project the project
	 */
	void attemptCVSUpdate(final IProject project, int syncMode) throws CoreException/*, InvocationTargetException, InterruptedException*/ {
		
		// If AutoSync is disabled for the project, we just leave it alone.
		if (!getAutoSync(project))
			return;
		
		// Create a callback action for successful update.
		// Failure to successfully update any AutoCVS project will
		// put us in disconnected mode.
		CVSOperations.OperationContext context = new CVSOperations.SimpleContext() {
			public void failure(Shell parent) {
				setFailedOperation(true);
			}
			public String getConsequenceMessage(int statusCode) {
				if (statusCode == CVSOperations.FILE_IN_THE_WAY)
					return getMessage("failed.update.fileintheway");
				else
					return getMessage("failed.update.generic");
			}
		};
		
		// Asynchronously execute the CVS operation
		CVSOperations.update(project, syncMode, context);
	}
	
	/**
	 * Automatically update all AutoCVS projects in the workspace. 
	 */
	private void autoUpdateWorkspaceProjects() throws CoreException, InvocationTargetException, InterruptedException {
		// Attempt a CVS update on projects with the AutoCVS nature
		IProject[] projectList = ResourcesPlugin.getWorkspace().getRoot().getProjects();
		Debug.print("Workspace contains " + projectList.length+" projects");
		for (int i = 0; i < projectList.length; ++i) {
			IProject project = projectList[i];
			if (hasAutoCVSNature(project))
				attemptCVSUpdate(project, CVSOperations.ASYNC);
			
			if (hasFailedOperation())
				break;
		}
	}
	
	private static boolean hasProjectNature(IProject project, String natureId) {
		try {
			return project.hasNature(natureId);
		} catch (CoreException e) {
			plugin.eventLog.logError("Exception getting project nature", e);
			return false;
		}
	}
	
	private void addProjectNature(IProject project, String natureId) throws CoreException {
		if (hasProjectNature(project, natureId))
			return;
		
		IProjectDescription projectDescription = project.getDescription();
		String[] ids = projectDescription.getNatureIds();
		String[] updateIds = new String[ids.length + 1];
		System.arraycopy(ids, 0, updateIds, 0, ids.length);
		updateIds[ids.length] = natureId;
		
		projectDescription.setNatureIds(updateIds);
		project.setDescription(projectDescription, null);
	}
	
	private void removeProjectNature(IProject project, String natureId) throws CoreException {
		if (!hasProjectNature(project, natureId))
			return;
		
		IProjectDescription projectDescription = project.getDescription();
		String[] ids =projectDescription.getNatureIds(); 
		String[] updateIds = new String[ids.length - 1];
		int count = 0;
		for (int i = 0; i < ids.length; ++i){
			if (!ids[i].equals(natureId))
				updateIds[count++] = ids[i];
		}
		
		projectDescription.setNatureIds(updateIds);
		project.setDescription(projectDescription, null);
	}
	
	/**
	 * Return whether or not the given project has the AutoCVS nature.
	 * @param project
	 * @return true if the project has the AutoCVS nature, false otherwise
	 */
	public static boolean hasAutoCVSNature(IProject project) {
		return hasProjectNature(project, AUTO_CVS_NATURE);
	}
	
	/**
	 * Add the AutoCVS nature to given project.
	 * @param project
	 */
	public void addAutoCVSNature(IProject project) throws CoreException {
		addProjectNature(project, AUTO_CVS_NATURE);
	}

	/**
	 * Remove the AutoCVS nature from given project.
	 * @param project
	 */
	public void removeAutoCVSNature(IProject project) throws CoreException {
		removeProjectNature(project, AUTO_CVS_NATURE);
	}

	public static boolean hasAutoSyncNature(IProject project) {
		return hasProjectNature(project, AUTO_SYNC_NATURE);
	}
	
	public void addAutoSyncNature(IProject project) throws CoreException {
		addProjectNature(project, AUTO_SYNC_NATURE);
	}
	
	public void removeAutoSyncNature(IProject project) throws CoreException {
		removeProjectNature(project, AUTO_SYNC_NATURE);
	}
	
	/**
	 * Get the single instance of the plugin object.
	 * @return the plugin object
	 */
	public static AutoCVSPlugin getPlugin() {
		return plugin;
	}
	
	/**
	 * Get the plugin id.
	 * @return the plugin id
	 */
	public String getId() {
		return getDescriptor().getUniqueIdentifier();
	}
	
	/**
	 * Get the event log.
	 * @return the event log
	 */
	public EventLog getEventLog() {
		return eventLog;
	}
	
	/**
	 * Convenience method for getting the handle of the active workbench window.
	 * @return the active workbench window, or null if we can't get it
	 */
	public static IWorkbenchWindow getActiveWorkbenchWindow() {
		IWorkbench workbench = PlatformUI.getWorkbench();
		if (workbench != null) {
			IWorkbenchWindow wwin = workbench.getActiveWorkbenchWindow();
			if (wwin == null)
				Debug.print("Could not get handle of active workbench window");
			return wwin;
		}
		return null;
	}
	
	/**
	 * Has any automatic CVS operation failed?
	 * @return true if an update has failed, false if not
	 */
	public boolean hasFailedOperation() {
		return failedOperation;
	}
	
	/**
	 * Set whether any automatic CVS operations have failed.
	 * @param failedOperation
	 */
	public void setFailedOperation(boolean failedOperation) {
		this.failedOperation = failedOperation;
	}
	
	/** Set the AutoSync property for given project. */
	public void setAutoSync(IProject project, boolean autoSync) throws CoreException {
		if (autoSync)
			addAutoSyncNature(project);
		else
			removeAutoSyncNature(project);
	}
	
	/** Get value of AutoSync property for given project. */
	public boolean getAutoSync(IProject project) throws CoreException {
		return hasAutoSyncNature(project);
	}
}
