/*
 * Created on Jan 13, 2004
 */
package edu.umd.cs.daveho.eclipse.autocvs;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.ui.dialogs.PropertyPage;

/**
 * A property page dialog for configuring the AutoCVS nature for projects.
 * This code adapted from the example in Chapter 19 of
 * <cite>Contributing to Eclipse</cite> by Gamma and Beck.
 * 
 * @author David Hovemeyer
 */
public class AutoCVSPropertyPage extends PropertyPage {
	
	// Fields
	private Button autoCVSButton;
	private Button autoSyncButton;
	
	private boolean origAutoCVS;
	private boolean origAutoSync;

	public AutoCVSPropertyPage() {
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.jface.preference.PreferencePage#createContents(org.eclipse.swt.widgets.Composite)
	 */
	protected Control createContents(Composite parent) {
		noDefaultAndApplyButton();
		boolean hasAutoCVSNature = false;
		boolean hasAutoSyncNature = false;
		try {
			hasAutoCVSNature = getProject().hasNature(AutoCVSPlugin.AUTO_CVS_NATURE);
			hasAutoSyncNature = getProject().hasNature(AutoCVSPlugin.AUTO_SYNC_NATURE);
			Debug.print("createContents: hasAutoCVSNature ==> " + hasAutoCVSNature);
		} catch (CoreException e) {
			AutoCVSPlugin.getPlugin().getEventLog().logError("Exception getting project nature", e);
		}
		Control control = addControl(parent, hasAutoCVSNature, hasAutoSyncNature);
		
		origAutoCVS = hasAutoCVSNature;
		origAutoSync = hasAutoSyncNature;
		
		return control;
	}
	
	public boolean performOk(){
		Debug.print("performOk() called");
		try {
			AutoCVSPlugin plugin = AutoCVSPlugin.getPlugin();
			boolean autoCVSEnabled =autoCVSButton.getSelection();
			boolean autoSyncEnabled = autoSyncButton.getSelection();
			
			Debug.print("\tAutoCVS ==> " + autoCVSEnabled);
			Debug.print("\tAutoSync ==> " + autoSyncEnabled);
			
			// Only reconfigure the project if the settings have
			// actually changed.
			if (autoCVSEnabled != origAutoCVS || autoSyncEnabled != origAutoSync) {
				if (autoCVSEnabled) {
					if (autoSyncEnabled) {
						// AutoSync has been turned on -
						// run a cvs update to make sure the project
						// starts out in sync with the repository.
						AutoCVSPlugin.getPlugin().attemptCVSUpdate(getProject(), CVSOperations.SYNC);
					}
					
					plugin.addAutoCVSNature(getProject());
					plugin.setAutoSync(getProject(), autoSyncEnabled);
				} else {
					plugin.removeAutoCVSNature(getProject());
					plugin.setAutoSync(getProject(), false);
				}
			}
		} catch (CoreException e){
			AutoCVSPlugin.getPlugin().getEventLog().logError("Exception getting project nature", e);
		}
		return true;
	}

	/**
	 * Get the project we're configuring properties of.
	 * @return the project
	 */
	private IProject getProject() {
		return (IProject) getElement();
	}

	/**
	 * Create the control used in the property page.
	 * @param parent the parent control
	 * @return the property page control
	 */
	private Control addControl(Composite parent, boolean hasAutoCVSNature, boolean hasAutoSyncNature) {
		Composite composite = new Composite(parent, SWT.NONE);
		GridLayout gridLayout = new GridLayout();
		gridLayout.numColumns = 1;
		composite.setLayout(gridLayout);
		
		// This sets the GridData for the composite itself
		GridData data = new GridData();
		data.verticalAlignment = GridData.FILL;
		data.horizontalAlignment = GridData.FILL;
		composite.setLayoutData(data);

		Font font = parent.getFont();
		Label label = new Label(composite, SWT.NONE);
		Label label2 = new Label(composite, SWT.NONE);
		label.setText("AutoCVS allows submission of projects, and (optionally)");
		label2.setText("keeps the project in sync with the repository automatically.");

		autoCVSButton = new Button(composite, SWT.CHECK);
		autoCVSButton.setText("Enable AutoCVS");
		autoCVSButton.setFont(font);
		autoCVSButton.setSelection(hasAutoCVSNature);
		
		// The selection listener ensures:
		// - If AutoCVS is checked, we also enable and check AutoSync,
		//		because AutoSync defaults to on.
		// - If AutoCVS is unchecked, AutoSync is disabled and unchecked.
		autoCVSButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent event) {
				if (autoCVSButton.getSelection()) {
					autoSyncButton.setSelection(true);
					autoSyncButton.setEnabled(true);
				} else {
					autoSyncButton.setSelection(false);
					autoSyncButton.setEnabled(false);
				}
			}
		});
		
		autoSyncButton = new Button(composite, SWT.CHECK);
		autoSyncButton.setText("Automatically keep project in sync with the repository");
		autoSyncButton.setFont(font);
		if (hasAutoCVSNature) {
			autoSyncButton.setSelection(hasAutoSyncNature);
		}
		
		// Indent the AutoSync button to show that it is
		// dependent on the status of the AutoCVS button.
		GridData syncButtonGridData = new GridData();
		syncButtonGridData.horizontalIndent = 15;
		autoSyncButton.setLayoutData(syncButtonGridData);
		
		return composite;
	}

}
