/*
 * Created on Jan 21, 2004
 */
package edu.umd.cs.daveho.eclipse.autocvs;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;

import org.eclipse.team.internal.ccvs.core.CVSTag;

/**
 * Meaningful tags associated with an AutoCVS project.
 * @author David Hovemeyer
 */
public class AutoCVSTagSet {
	private static final String UMDCS_PREFIX = "umdcs_";
	private static final String ASSIGNED_PREFIX = UMDCS_PREFIX + "assigned_";
	private static final String SUBMITTED_PREFIX = UMDCS_PREFIX + "submitted_";
	private static final String CLOSED_PREFIX = UMDCS_PREFIX + "closed_";

	/**
	 * Map of project number (Integer) to ProjectInfo.
	 */
	private TreeMap projectInfoMap = new TreeMap();

	/**
	 * Information about a project.
	 * Contains list of assigned projects, which ones have been closed,
	 * and the current highest submission number for each project.
	 * 
	 * @author David Hovemeyer
	 */
	private static class ProjectInfo {
		private int projectNumber;
		private boolean open = true;
		private int submission = 0;

		/**
		 * Constructor.
		 * 
		 * @param projectNumber the project number
		 */
		ProjectInfo(int projectNumber) {
			this.projectNumber = projectNumber;
		}
		
		/**
		 * Set whether the project is open.
		 * @param open whether the project is open
		 */
		public void setOpen(boolean open) {
			this.open = open;
		}
		
		/**
		 * Return whether the project is open.
		 * @return true if the project is open, false if closed
		 */
		public boolean isOpen() {
			return open;
		}
		
		/**
		 * Set the current highest existing submission number.
		 * @param submission the highest submission number
		 */
		public void setSubmission(int submission) {
			this.submission = submission;
		}
		
		/**
		 * Get the next CVS tag to use to submit a version of the project.
		 * @return the tag
		 */
		public String nextSubmissionTag() {
			StringBuffer buf = new StringBuffer();
			buf.append(SUBMITTED_PREFIX);
			buf.append(projectNumber);
			buf.append('_');
			buf.append(++submission);
			
			return buf.toString();
		}

		/**
		 * Get the current highest submission number for this project.
		 * @return the current highest submission
		 */
		public int getSubmission() {
			return submission;
		}
	}
	
	/**
	 * Initialize the object based on the list of CVS tags.
	 * 
	 * @param tagList the list of CVS tags
	 */
	public void init(CVSTag[] tagList) {
		// We support the following kinds of tags:
		// umdcs_assigned_<n>
		//		project <n> was assigned
		// umdcs_submitted_<n>_<v>
		//		student submitted version <v> of project <n>
		// umdcs_closed_<n>
		//		project <n> is closed (no further submissions accepted)
		
		// First pass: find assigned projects (including those that are closed)
		for (int i = 0; i < tagList.length; i++) {
			CVSTag tag = tagList[i];
			String name = tag.getName();
			
			Debug.print("Found project tag " + name);
			
			if (name.startsWith(ASSIGNED_PREFIX)) {
				name = name.substring(ASSIGNED_PREFIX.length());
				try {
					int projectNumber = Integer.parseInt(name);
					getProjectInfo(projectNumber);
					Debug.print("Project " + projectNumber + " has been assigned");
				} catch (NumberFormatException e) {
					// Ignored
				}
			} else if (name.startsWith(CLOSED_PREFIX)) {
				name = name.substring(CLOSED_PREFIX.length());
				try {
					int projectNumber = Integer.parseInt(name);
					ProjectInfo info = getProjectInfo(projectNumber);
					info.setOpen(false);
				} catch (NumberFormatException e) {
					// Ignored
				}
			}
		}
		
		// Second pass looks for submitted versions of each project
		for (int i = 0; i < tagList.length; i++) {
			CVSTag tag = tagList[i];
			String name = tag.getName();
			
			if(name.startsWith(SUBMITTED_PREFIX)) {
				name = name.substring(SUBMITTED_PREFIX.length());
				try {
					int sep;
					
					sep = name.indexOf('_');
					if (sep > 0) {
						int projectNumber = Integer.parseInt(name.substring(0, sep));
						name = name.substring(sep + 1);
						int submission = Integer.parseInt(name);
						
						ProjectInfo info = getProjectInfo(projectNumber);
						if (submission > info.getSubmission())
							info.setSubmission(submission);
					}
				} catch (NumberFormatException e) {
					// Ignore
				}
			}
		}
	}
	
	/**
	 * Get list of numbers of projects that are both assigned an opened.
	 * These are candidates for submission.
	 * If there are multiple projects, they will be sorted
	 * in ascending order.
	 * @return the list of numbers of open projects
	 */
	public int[] getOpenProjects() {
		ArrayList openList = new ArrayList();
		for (Iterator i =projectInfoMap.entrySet().iterator(); i.hasNext(); ) {
			Map.Entry entry = (Map.Entry) i.next();
			Integer key = (Integer) entry.getKey();
			ProjectInfo info = (ProjectInfo) entry.getValue();
			if (info.isOpen())
				openList.add(key);
		}
		int[] result = new int[openList.size()];
		int count = 0;
		for (Iterator iter = openList.iterator(); iter.hasNext();) {
			Integer element = (Integer) iter.next();
			result[count++] = element.intValue();
		}
		return result;
	}
	
	public String getNextSubmissionTag(int projectNumber) {
		ProjectInfo info = getProjectInfo(projectNumber);
		return info.nextSubmissionTag();
	}
	
	private ProjectInfo getProjectInfo(int projectNumber) {
		Integer key = new Integer(projectNumber);
		ProjectInfo info = (ProjectInfo) projectInfoMap.get(key);
		if (info == null) {
			info = new ProjectInfo(projectNumber);
			projectInfoMap.put(key, info);
		}
		return info;
	}
}
