/*
 * Created on Jan 22, 2004
 */
package edu.umd.cs.daveho.eclipse.autocvs;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.team.core.RepositoryProvider;
import org.eclipse.team.core.TeamException;
import org.eclipse.team.internal.ccvs.core.CVSException;
import org.eclipse.team.internal.ccvs.core.CVSTag;
import org.eclipse.team.internal.ccvs.core.CVSTeamProvider;
import org.eclipse.team.internal.ccvs.core.ICVSRemoteFile;
import org.eclipse.team.internal.ccvs.core.ICVSRemoteFolder;
import org.eclipse.team.internal.ccvs.core.ICVSRemoteResource;
import org.eclipse.team.internal.ccvs.core.ICVSResource;
import org.eclipse.team.internal.ccvs.core.ILogEntry;
import org.eclipse.team.internal.ccvs.core.client.Command;
import org.eclipse.team.internal.ccvs.core.resources.CVSWorkspaceRoot;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.actions.WorkspaceModifyOperation;

/**
 * Facade for CVS operations.
 * @author David Hovemeyer
 */
public class CVSOperations {
	// Operation failure reasons
	private static final int FAILURE_REASON_START = -100001; // Don't conflict with CVS status codes
	public static final int UNKNOWN_FAILURE = FAILURE_REASON_START + -1;
	public static final int FAILED_UP_TO_DATE_CHECK = FAILURE_REASON_START + -2;
	public static final int UNRESOLVED_MERGE_CONFLICT = FAILURE_REASON_START + -3;
	public static final int FILE_IN_THE_WAY = FAILURE_REASON_START + -4;
	
	/*
	 * Uniform interface for a CVS operation ready to execute.
	 * It will be executed in the context of a WorkspaceModifyOperation,
	 * and thrown errors will be handled (and displayed to the user)
	 * to the best of our ability.
	 */
	private interface Operation {
		/**
		 * Execute the operation.
		 * @param parent parent Shell, null if none
		 * @param monitor progress monitor
		 * @throws CoreException
		 * @throws InvocationTargetException
		 */
		public void execute(Shell parent, IProgressMonitor monitor) throws CoreException, InvocationTargetException/*, InterruptedException*/;

		/**
		 * Get descriptive name for the operation.
		 * This will be the task name used for the progress monitor,
		 * and it will also be used if an error dialog is required.
		 * 
		 * @return the operation name
		 */
		public String getOperationName();

		/**
		 * Get descriptive message describing the task performed by the operation.
		 * @return the message describing the operation
		 */
		public String getMessage();
	}
	
	/**
	 * Hook for indication of the success or failure of a CVS operation.
	 * Giving the same OperationContext to two successive asynchronous
	 * operations can have the effect of ensuring that each operation
	 * proceeds only if the previous operation was successful. 
	 */
	public interface OperationContext {
		public void success(Shell parent);
		public void failure(Shell parent);
		public boolean operationMayProceed();
		public String getConsequenceMessage(int statusCode);
	}
	
	/**
	 * Simple implementation of OperationContext that just records the
	 * success or failure of a CVS operation as a boolean.
	 * If the same  SimpleContext is used for multiple asynchronous
	 * operations, it will ensure that each operation proceeds only
	 * if the previous operation was successful.
	 */
	public static class SimpleContext implements OperationContext {
		public boolean success = true;
		
		public void success(Shell parent) {
			this.success = true;
		}

		public void failure(Shell parent) {
			this.success = false;
		}

		public boolean operationMayProceed() {
			return success;
		}

		public String getConsequenceMessage(int statusCode) {
			return null;
		}
	}
	
	// Constants
	/** Execute CVS operation synchronously (in calling thread). */
	public static final int SYNC = 0;

	/** Execute CVS operation asynchronously (scheduling it in the GUI dispatch thread). */
	public static final int ASYNC = 1;
	
	/**Execute CVS operation in a top level dialog. */
	public static final int IN_DIALOG = 0;
	
	/** Execute CVS operation in the status bar of the active workspace window. */
	public static final int IN_STATUS_BAR = 1;
	
	private static void validateSyncMode(int syncMode) {
		if (syncMode != SYNC && syncMode != ASYNC)
			throw new IllegalStateException("Invalid sync mode for cvs operation" +syncMode);
	}
	
	/**
	 * Update given project to bring it up to date with the latest repository version.
	 * @param project the project to update
	 * @param callback an action to perform after the update
	 * @param syncMode either CVSOperations.SYNC or CVSOperations.ASYNC
	 * @throws TeamException
	 */
	public static void update(final IProject project, int syncMode) {
		update(project, syncMode, null);
	}
	
	/**
	 * Update given project to bring it up to date with the latest repository version.
	 * @param project the project to update
	 * @param callback an action to perform after the update
	 * @param syncMode either CVSOperations.SYNC or CVSOperations.ASYNC
	 * @param context action to perform following successful update
	 * @throws TeamException
	 */
	public static void update(final IProject project, int syncMode, final OperationContext context) {
		validateSyncMode(syncMode);
		
		final Operation runnable = new Operation() {
			public void execute(Shell parent, IProgressMonitor monitor) throws CoreException, InvocationTargetException {
				Debug.print("Performing automatic CVS  update of project " + project.getName() +
						" @" + System.identityHashCode(project));
				Debug.print("Project class is " + project.getClass().getName());

				final CVSTeamProvider provider = CVSOperations.getProvider(project);
				if (provider == null)
					throw new CVSException("Internal error: No CVS provider for project " + project.getName());
				
				Debug.print("Running the CVS update!");
				AutoCVSPlugin.getPlugin().getEventLog().logMessage("Auto cvs update for project " + project.getName());
				provider.update(
						new IResource[]{ project },
						Command.NO_LOCAL_OPTIONS,
						null, // cvsTag
						true, // createBackups
						monitor
				);
				Debug.print("Done with CVS update");
			}

			public String getOperationName() {
				return "Automatic CVS Update";
			}

			public String getMessage() {
				return "Update local files to synchronize with repository";
			}
		};
		
		// Schedule the update for execution in the GUI dispatch thread
		asyncExec(runnable, IN_STATUS_BAR, context);
	}
	
	/**
	 * Commit resources to the CVS repository.
	 * 
	 * @param resourceList list of resources to commit
	 * @param syncMode SYNC or ASYNC
	 * @param context the OperationContext
	 */
	public static void commit(final IResource[] resourceList, int syncMode, OperationContext context) throws InvocationTargetException, InterruptedException {
		if (resourceList.length == 0)
			return; // Nothing to commit
		
		Operation runnable = new Operation() {
			public void execute(Shell parent, IProgressMonitor monitor) throws CoreException, InvocationTargetException  {
				monitor.setTaskName("Automatic CVS commit");
				
				// Get the single repository provider for all resources
				CVSTeamProvider provider = getSingleProvider(resourceList);
				if (provider == null)
					throw new CVSException("Internal error: Couldn't get single provider for committed resources");
				
				// Do the commit.
				AutoCVSPlugin.getPlugin().getEventLog().logMessage("Committing resources: " +
						resourceArrayToString(resourceList));
				provider.checkin(resourceList, IResource.DEPTH_ZERO, monitor);
			}

			public String getOperationName() {
				return "Automatic CVS Commit";
			}

			public String getMessage() {
				return "Commit modifications to local files to the repository";
			}
		};
		
		// Schedule the commit.
		exec(runnable, syncMode, IN_STATUS_BAR, context);
	}
	
	private static class TagSetHolder {
		public CVSTag[] tagSet;
	}
	
	/**
	 * Get list of CVS tags for given resource.
	 * This operation must be performed synchronously.
	 * 
	 * @param resource the resource
	 * @param how either CVSOperations.IN_DIALOG or CVSOperations.IN_STATUS_BAR
	 * @return list of cvs tags, or null if couldn't fetch the tags
	 * @throws TeamException
	 */
	public static CVSTag[] fetchTags(final Shell parent, final IResource resource, int how) throws TeamException, InvocationTargetException, InterruptedException {
		final TagSetHolder tagSetHolder = new TagSetHolder();
		
		Operation runnable = new Operation() {
			public void execute(Shell parent, IProgressMonitor monitor) throws CoreException, InvocationTargetException {
				if (!AutoCVSPlugin.isCVSManaged(resource))
					throw new CVSException("Resource " + resource.getName() + " is not CVS managed");
				
				// Get the ICVSRemoteFile
				ICVSResource projectCVSResource =
					CVSWorkspaceRoot.getCVSResourceFor(resource);
				ICVSRemoteResource projectRemoteResource =
					CVSWorkspaceRoot.getRemoteResourceFor(projectCVSResource);
				if (!(projectRemoteResource instanceof ICVSRemoteFile))
					throw new CVSException("Resource " + resource.getName() + " is not a CVS-managed file");
				ICVSRemoteFile remoteFile = (ICVSRemoteFile) projectRemoteResource;

				ArrayList tagList = new ArrayList();

				// Fetch log entries for the file.
				ILogEntry[] logEntryList = remoteFile.getLogEntries(monitor);
				for (int i = 0; i < logEntryList.length; i++) {
					ILogEntry logEntry = logEntryList[i];
					CVSTag[] entryTagList = logEntry.getTags();
					for (int j = 0; j < entryTagList.length; j++) {
						tagList.add(entryTagList[j]);
					}
				}
				Debug.print("Found " + tagList.size() + " tags for project");
				
				// Successfully got the tags.  Put them in the TagSetHolder
				tagSetHolder.tagSet = (CVSTag[]) tagList.toArray(new CVSTag[0]);
			}

			public String getOperationName() {
				return "Fetch CVS Tags for Project";
			}

			public String getMessage() {
				return "Fetch tags describing assignments from repository";
			}
		};
		
		syncExec(runnable, how, null);
		
		return tagSetHolder.tagSet;
	}
	
	/**
	 * Tag all managed resources in given project.
	 * The tag is a plain revision tag, not a branch tag.
	 * @param project the project to tag
	 * @param tagName name of the tag
	 */
	public static void tagProject(final IProject project, final String tagName, int syncMode) throws CVSException, InvocationTargetException, InterruptedException {
		validateSyncMode(syncMode);
		
		Operation runnable = new Operation(){
			public void execute(Shell parent, IProgressMonitor monitor) throws CoreException, InvocationTargetException {
				ICVSResource cvsResource = CVSWorkspaceRoot.getCVSResourceFor(project);
				if (cvsResource == null)
					throw new CVSException("Couldn't get ICVSResource for " + project.getName());
				ICVSRemoteResource remoteResource = CVSWorkspaceRoot.getRemoteResourceFor(cvsResource);
				if (remoteResource == null)
					throw new CVSException("Couldn't get ICVSRemoteResource for " + project.getName());
				
				if (!(remoteResource instanceof ICVSRemoteFolder))
					throw new CVSException("Couldn't get ICVSRemoteFolder for " + project.getName());
				
				CVSTag tag = new CVSTag(tagName, CVSTag.VERSION);
				
				monitor.setTaskName("Tagging project files for submission");
				AutoCVSPlugin.getPlugin().getEventLog().logMessage("Tagging project " +
						project.getName() + " with tag " + tag.getName());
				Debug.print("Adding tag " +tagName + " to project");
				remoteResource.tag(tag, Command.NO_LOCAL_OPTIONS, monitor);
			}

			public String getOperationName() {
				return "CVS Tag Project Files";
			}

			public String getMessage() {
				return "Tag files in project to mark them as submitted";
			}
		};
		
		exec(runnable, syncMode, IN_STATUS_BAR, null);
	}

	/**
	 * Add given resources to repository.
	 * @param resources list of resources to add
	 * @param syncMode SYNC or ASYNC
	 * @param context the OperationContext
	 */
	public static void add(final IResource[] resources, int syncMode, OperationContext context) {
		Operation runnable = new Operation() {
			public void execute(Shell parent, IProgressMonitor monitor) throws CoreException, InvocationTargetException {
				if (resources.length == 0) {
					// Nothing to add
					return;
				}
				
				// Find the project.
				// All added resources must belong to the same project.
				IProject project = null;
				for (int i = 0; i < resources.length; i++) {
					IProject rProject = resources[i].getProject();
					if (rProject == null)
						throw new IllegalStateException("Adding a resource with no project!");
					if (project == null)
						project = rProject;
					else if (project != rProject)
						throw new CVSException("Internal error: Adding resources from multiple projects");
				}
				
				// Use the project's resource provider to add the resources
				CVSTeamProvider provider = getSingleProvider(new IResource[]{project});
				if (provider == null)
					throw new CVSException("Internal error: Could not get resource provider for added resources");
				
				AutoCVSPlugin.getPlugin().getEventLog().logMessage("Adding resources: " +
						resourceArrayToString(resources));
				provider.add(resources, IResource.DEPTH_ZERO, monitor);
			}

			public String getOperationName() {
				return "Automatic CVS Add";
			}

			public String getMessage() {
				return "Add new local files to the repository";
			}
		};
		
		exec(runnable, syncMode, IN_STATUS_BAR, context);
	}

	private static String resourceArrayToString(IResource[] resourceList) {
		StringBuffer buf = new StringBuffer();
		for (int i = 0; i < resourceList.length; i++) {
			buf.append(resourceList[i].getName());
			if (i < resourceList.length - 1)
				buf.append(',');
		}
		return buf.toString();
	}
	
	/**
	 * Get CVSTeamProvider which is the repository provider for
	 * given resource. Returns null if the resource has no provider,
	 * or if the provider is not CVS.
	 * 
	 * @param resource the resource
	 * @return the CVSRepositoryProvider for the project, or null if the provider is not CVS
	 */
	public static CVSTeamProvider getProvider(IResource resource) {
		RepositoryProvider provider = RepositoryProvider.getProvider(resource.getProject());
		if (provider == null || !(provider instanceof CVSTeamProvider))
			return null;
		return (CVSTeamProvider) provider;
	}

	/**
	 * Get the single CVSTeamProvider managing all of a list of resources.
	 * Returns null if any of the resources are unmanaged, or if
	 * there are multiple providers.
	 * 
	 * @param resourceList list of resources
	 * @return the single CVSTeamProvider, or null if there is no single provider
	 *    for all of the resources
	 */
	public static CVSTeamProvider getSingleProvider(IResource[] resourceList) {
		CVSTeamProvider provider = null;
		for (int i = 0; i < resourceList.length; ++i) {
			IResource resource = resourceList[i];
			
			CVSTeamProvider resProvider = getProvider(resource);
			if (resProvider == null) {
				AutoCVSPlugin.getPlugin().getEventLog().logError("Resource " + resource.getName() + " has no RepositoryProvider");
				return null;
			}
			
			if (provider == null)
				provider = resProvider;
			else if (provider != resProvider) {
				AutoCVSPlugin.getPlugin().getEventLog().logError("Multiple repository providers for resources");
				return null;
			}
		}
		return provider;
	}
	
	private static void exec(Operation operation, int syncMode, int how, OperationContext context) {
		if (syncMode == SYNC) {
			syncExec(operation, how, context);
		} else if (syncMode == ASYNC) {
			asyncExec(operation, how, context);
		} else
			throw new IllegalStateException();
	}

	/**
	 * Run a CVS operation asynchronously in a progress dialog
	 * or workbench status bar.
	 * @param runnable the operation
	 * @param how IN_STATUS_BAR to request running using the status bar,
	 *   IN_DIALOG to request running in a progress dialog
	 * @param context the OperationContext
	 */
	private static void asyncExec(final Operation runnable, final int how, final OperationContext context) {
		Display.getDefault().asyncExec(new Runnable(){
			public void run() {
				syncExec(runnable, how, context);
			}
		});
	}
	
	/**
	 * Run a CVS operation synchronously (in the context of the current thread),
	 * in a progress dialog or workbench status bar.  Displays an error
	 * dialog if the CVS operation does not complete successfully.
	 * 
	 * @param runnable the operation
	 * @param how IN_STATUS_BAR to request running using the status bar,
	 *   IN_DIALOG to request running in a progress dialog
	 * @param context the OperationContext
	 * @throws InvocationTargetException
	 * @throws InterruptedException
	 */
	private static void syncExec(final Operation cvsOp, final int how, final OperationContext context) {
		// The Operation context, if present, may veto this entire operation.
		// This generally means that a previous asynchronously scheduled operation
		// failed.
		if (context != null && !context.operationMayProceed())
			return;
		
		final IWorkbenchWindow wwin = AutoCVSPlugin.getActiveWorkbenchWindow();
		final Shell parent = (wwin != null) ? wwin.getShell() : null;
		
		// Wrap the Operation in a WorkspaceModifyOperation
		WorkspaceModifyOperation runnable = new WorkspaceModifyOperation() {
			protected void execute(IProgressMonitor monitor) throws CoreException, InvocationTargetException, InterruptedException {
				try {
					monitor.setTaskName(cvsOp.getOperationName());
					cvsOp.execute(parent, monitor);
				} finally {
					monitor.done();
				}
			}
		};

		try {
			if (how == IN_STATUS_BAR) {
				if (wwin != null) {
					wwin.run(false, true, runnable);
					if (context != null)
						context.success(parent);
					return;
				}
			}
			
			// A modal progress dialog is the fall back for the
			// case where we can't get a handle on the workbench window.
			ProgressMonitorDialog dialog = new ProgressMonitorDialog(null);
			dialog.run(false, true, runnable);
			context.success(parent);
			return;
		} catch (Throwable e) {
			//Failed operation!
			if (context != null)
				context.failure(parent);
			
			// Attempt to figure out the reason for the failure from the exception.
			int statusCode = parseStatus(e);
			
			// Build a detailed message describing the failure and its consequences.
			StringBuffer msg = new StringBuffer();
			msg.append("The following operation failed: " + cvsOp.getMessage());
			String consequenceMessage = null;
			if (context != null)
				consequenceMessage = context.getConsequenceMessage(statusCode);
			if (consequenceMessage != null) {
				msg.append("\n\n");
				msg.append(consequenceMessage);
			}
			
			if (AutoCVSPlugin.getPlugin().hasFailedOperation()) {
				// Add the generic message about how automatic CVS commands
				// are disabled.
				msg.append("\n\n");
				msg.append(AutoCVSPlugin.getMessage("error.offline"));
			}
			
			// Alert the user.
			Dialogs.errorDialog(
					parent,
					"Warning: " + cvsOp.getOperationName() + " failed",
					msg.toString(),
					e
			);
			context.failure(parent);
		}
	}

	/**
	 * Ad-hoc parsing of status messages to try to glean information about
	 * an error.
	 * @param e an Exception thrown by a CVS operation
	 * @return return a status code - hopefully, one of the failure reason codes defined in this class
	 */
	private static int parseStatus(Throwable e) {
		int statusCode = 0;
		
		// Try to get status
		Throwable cause = e;
		if (e instanceof InvocationTargetException)
			cause = ((InvocationTargetException)e).getTargetException();
		
		IStatus status = null;
		if (cause instanceof CoreException){
			CoreException ce = (CoreException) cause;
			status = ce.getStatus();
			Debug.print("status code = " + status.getCode());
			statusCode = status.getCode(); // generally useless, but who knows?
			
			List statusMessageList = getStatusMessages(status);
			for (Iterator iter = statusMessageList.iterator();
				iter.hasNext();
				) {
				String element = (String) iter.next();
				if (element.indexOf("Up-to-date check") >= 0) {
					statusCode = FAILED_UP_TO_DATE_CHECK;
					break;
				} else if (element.indexOf("had a conflict") >= 0) {
					statusCode = UNRESOLVED_MERGE_CONFLICT;
					break;
				} else if (element.indexOf("in the way") >= 0 ||
						element.indexOf("added independently") >= 0 ||
						element.indexOf("already exists") >= 0) {
					statusCode = FILE_IN_THE_WAY;
					break;
				}
			}
		}
		Debug.print("Final status code = " + statusCode);
		
		return statusCode;
	}
	
	private static List getStatusMessages(IStatus status) {
		ArrayList list = new ArrayList();
		list.add(status.getMessage());
		IStatus childList[] = status.getChildren();
		for (int i = 0; i < childList.length; i++) {
			list.add(childList[i].getMessage());
		}
		return list;
	}
}
