/*
 * Created on Feb 2, 2004
 */
package edu.umd.cs.daveho.eclipse.autocvs;

import java.lang.reflect.InvocationTargetException;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Shell;

/**
 * Convenience methods for error and success dialogs.
 * In addition to displaying a modal dialog, these methods
 * log events to the AutoCVS console.
 * 
 * @author David Hovemeyer
 */
public class Dialogs {

	/**
	 * Show an error dialog caused by receiving an exception.
	 * @param parent parent Shell, null if none
	 * @param title the dialog title
	 * @param message description of the error, if one can't be found automatically
	 * @param e the exception
	 */
	public static void errorDialog(Shell parent, String title, String message, Throwable e) {
		// Log the error.
		AutoCVSPlugin.getPlugin().getEventLog().logError(title + ": " + message, e);
		
		// If e is an InvocationTargetException, then the wrapped
		// exception is much more likely to have useful information in it
		// than the outer exception.
		if (e instanceof InvocationTargetException)
			e = ((InvocationTargetException) e).getTargetException();

		// Create IStatus object describing the error.
		// Hopefully we're looking at a CoreException where the
		// status object is already present.  If not, we just
		// use in a generic message to describe the error.
		IStatus status;
		if (e instanceof CoreException) {
			final IStatus exceptionStatus = ((CoreException)e).getStatus();
			// Reported status should have a severity less than errors
			if (exceptionStatus.getSeverity() != IStatus.ERROR) {
				status = exceptionStatus;
			} else {
				// Wrap the IStatus to turn it into a warning
				status = new IStatus() {
					public IStatus[] getChildren() {return exceptionStatus.getChildren();}
					public int getCode() {return exceptionStatus.getCode();}
					public Throwable getException() {return exceptionStatus.getException();}
					public String getMessage() {return exceptionStatus.getMessage();}
					public String getPlugin() {return exceptionStatus.getPlugin();}
					public int getSeverity() {return IStatus.WARNING;}
					public boolean isMultiStatus() {return exceptionStatus.isMultiStatus();}
					public boolean isOK() {return exceptionStatus.isOK();}
					public boolean matches(int severityMask) {return exceptionStatus.matches(severityMask);}
				};
			}
		} else {
			status = new Status(IStatus.WARNING,
					AutoCVSPlugin.getPlugin().getDescriptor().getUniqueIdentifier(),
					0, // TODO: need a code here?
					message, e);
		}

		// Create a nice obnoxious dialog to inform the user.
		new ErrorDialog(
				parent,
				title,
				message,
				status,
				IStatus.CANCEL|IStatus.ERROR|IStatus.INFO|IStatus.WARNING
		).open();
	}
	

	/**
	 * Convenience method for displaying an error dialog.
	 * @param parent parent Shell, null if none
	 * @param title dialog title
	 * @param message message to be displayed
	 * @param reason reason for the error 
	 */
	public static void errorDialog(Shell parent, String title, String message, String reason) {
		// Log the error.
		AutoCVSPlugin.getPlugin().getEventLog().logError(title + ": " + message);
	
		IStatus status =new Status(IStatus.WARNING, AutoCVSPlugin.getPlugin().getId(), 
				0, reason, null); 
		new ErrorDialog(
				parent,
				title,
				message,
				status,
				IStatus.ERROR).open();
	}


	/**
	 * Dialog to show the successful completion of a task.
	 * @param parent parent Shell, null if none
	 * @param title dialog title
	 * @param message message to be displayed
	 */
	public static void okDialog(Shell parent, String title, String message) {
		// Log the event
		AutoCVSPlugin.getPlugin().getEventLog().logMessage(title + ": " + message);
		
		// Tagging appears to have succeeded.
		// Inform the user.
		new MessageDialog(
				parent,
				title,
				null,
				message,
				MessageDialog.OK,
				new String[]{"OK"},
				0).open();
	}
}
