/*
 * Created on Jan 21, 2004
 */
package edu.umd.cs.daveho.eclipse.autocvs;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectNature;

import org.eclipse.core.runtime.CoreException;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceVisitor;

import org.eclipse.jface.action.IAction;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.team.core.TeamException;
import org.eclipse.team.internal.ccvs.core.CVSException;
import org.eclipse.team.internal.ccvs.core.CVSTag;

import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorReference;

import org.eclipse.ui.IObjectActionDelegate;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;

/**
 * Action to submit an AutoCVS project.
 * The user picks an assigned open project from the AutoCVSTagSet
 * for the project.  Then we add a new "submitted" tag to the project
 * for that project number.  The submission numbers increase
 * as the project is submitted multiple times, so the highest submission
 * number is the most recent.
 * 
 * @author David Hovemeyer
 */
public class SubmitProjectAction implements IObjectActionDelegate {
	// Fields
	private ISelection selection;

	/* (non-Javadoc)
	 * @see org.eclipse.ui.IObjectActionDelegate#setActivePart(org.eclipse.jface.action.IAction, org.eclipse.ui.IWorkbenchPart)
	 */
	public void setActivePart(IAction action, IWorkbenchPart targetPart) {
		// TODO Auto-generated method stub

	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.IActionDelegate#run(org.eclipse.jface.action.IAction)
	 */
	public void run(IAction action) {
		Debug.print("Submit project action invoked");
		if (!(selection instanceof IStructuredSelection))
			return;
		IStructuredSelection structured = (IStructuredSelection) selection;
		Object obj = structured.getFirstElement();
		Debug.print("Selection object is a " + obj.getClass().getName() + " @" +
				System.identityHashCode(obj));

		// Get the IProject handle.
		if (!(obj instanceof IProjectNature)) {
			Dialogs.errorDialog(null, "Warning: Selection is Invalid",
					"Invalid submit action: You have selected an object that is not a Project. Please select a Project and try again.",
					"Object selected is not a Project");
			return;
		}
		IProject project = ((IProjectNature)obj).getProject();
		Debug.print("Got the IProject for the submit action @" + System.identityHashCode(project));
		
		// Run the dialog to actually submit the project
		runSubmitDialog(action, project);
	}
	
	/**
	 * Run the submit dialog for given project.
	 * @param action the "submit project" action
	 * @param project the project
	 * @param tagSet the tag set representing assigned, submitted, and closed projects
	 */
	private void runSubmitDialog(IAction action, IProject project) {
		IWorkbench workbench = PlatformUI.getWorkbench();
		if (workbench == null) {
			Dialogs.errorDialog(null, "Warning: project submission failed",
					"Could not submit project",
					"Internal error: Can't get workbench");
			return;
		}
		IWorkbenchWindow wwin = workbench.getActiveWorkbenchWindow();
		if (wwin == null) {
			Dialogs.errorDialog(null, "Error submitting project",
					"Could not submit project",
					"Internal error: Can't get workbench window");
			return;
		}
		
		// Shell to use as parent of dialogs.
		Shell parent = wwin.getShell();
		
		try {
			// Get tags for project, so we know what projects are eligible for submission
			AutoCVSTagSet tagSet = getTagSet(parent, project);
			
			// Note: if there are no projects available for submission,
			// then we skip prompting the user to save unsaved files.
			// This allows SubmitDialog to perform the task of notifying
			// the user that there is nothing to submit without having
			// to first bother the user about unsaved files.
			if (tagSet.getOpenProjects().length > 0) {
				// If the project contains unsaved files, prompt user to save them.
				// If he/she elects not to save them, then we won't submit the project.
				if (!saveDirtyEditors(project, workbench))
					return;
			}
			
			// Special case: if AutoSync has been disabled because of a failed
			// cvs update or because it was explicitly disabled,
			// but the user chooses to submit anyway, we need
			// to add/commit all non-ignored resources in order to bring the
			// project in sync with the repository. If we didn't, unmanaged files
			// wouldn't get tagged, and some tags would correspond to stale
			// file versions.
			if (AutoCVSPlugin.getPlugin().hasFailedOperation() || !AutoCVSPlugin.getPlugin().getAutoSync(project)) {
				AutoCVSPlugin.getPlugin().getEventLog().logMessage(
						"Syncing project for disconnected submit operation for project " +
						project.getName());

				// Completion callback to record the success or failure of the
				// add and submit actions.
				CVSOperations.SimpleContext completion = new CVSOperations.SimpleContext();

				// Add all non-ignored files in the project
				Set resourceSet = getProjectResources(project);
				ArrayList addedFiles = new ArrayList();
				for (Iterator iter = resourceSet.iterator(); iter.hasNext();) {
					IResource resource = (IResource) iter.next();
					if (resource instanceof IFile) {
						IFile file = (IFile) resource;
						if (!AutoCVSPlugin.isCVSIgnored(file) && !AutoCVSPlugin.isCVSManaged(file)) {
							addedFiles.add(file);
						}
					}
				}
				IResource[] files = (IResource[]) addedFiles.toArray(new IResource[0]);
				CVSOperations.add(files, CVSOperations.SYNC, completion);
				if (!completion.success)
					return; // User has seen failure dialog for failed cvs add
				
				// Commit the project to ensure all files are in sync with
				// the repository.
				CVSOperations.commit(new IResource[]{project}, CVSOperations.SYNC, completion);
				if (!completion.success)
					return; // User has seen failure dialog for failed cvs commit
			}
			
			// Run the submit dialog
			SubmitDialog dialog = new SubmitDialog(parent, tagSet);
			dialog.open();
			int submitProject = dialog.getResult();
			if (submitProject >= 0) {
				// Project selected: tag it
				tagProject(project, tagSet, submitProject);
				
				// Inform user of success
				Dialogs.okDialog(
						parent,
						"Project submission successful",
						"Project " + submitProject + " was submitted successfully");
			}
		} catch (Exception e) {
			Dialogs.errorDialog(
					parent, 
					"Warning: Project submission failed", 
					"Project submission failed", 
					e);
		}
		
	}

	/**
	 * Tag a project for submission.
	 * @param project the project
	 * @param tagSet the project's tag set
	 * @param submitProject number of project to submit
	 * @throws CVSException
	 * @throws InvocationTargetException
	 * @throws InterruptedException
	 */
	private void tagProject(IProject project, AutoCVSTagSet tagSet, int submitProject) throws CVSException, InvocationTargetException, InterruptedException {
		// A project was selected to submit.
		// Run the tag operation.
		String tagName = tagSet.getNextSubmissionTag(submitProject);
		Debug.print("Next submission tag: " + tagName);
		// Execute the tag operation (in this thread) and wait for it to complete.
		CVSOperations.tagProject(project, tagName, CVSOperations.SYNC);
	}

	/**
	 * Get populated tag set for given project.
	 * @param parent parent Shell, null if none
	 * @param project the project
	 * @return the populated tag set
	 */
	private AutoCVSTagSet getTagSet(Shell parent, IProject project) throws TeamException, InvocationTargetException, InterruptedException {
		// TODO Auto-generated method stub
		
		// Get the tag set, so we know which projects are available to submit
		AutoCVSTagSet tagSet = new AutoCVSTagSet();
		
		// Find the .project file IResource
		IResource projectFile = project.findMember(".project");
		if (projectFile == null) {
			Dialogs.errorDialog(parent, "Warning: project submission failed",
					"Could not submit project",
					"Can not find .project file for project");
			return null;
		}
		Debug.print("Found the .project file for the project!");
		
		// Get the tags for the .project file
		CVSTag[] tagList = CVSOperations.fetchTags(parent, projectFile, CVSOperations.IN_STATUS_BAR);
		
		// Create the AutoCVSTagSet for the project.
		// This will initialize the information about assigned and submitted projects.
		tagSet.init(tagList);
		
		return tagSet;
	}

	/**
	 * Save all dirty editors in given project.
	 * @param project the project
	 * @param workbench the workbench
	 * @return true if no editors are dirty, or if all dirty editors were successfully saved
	 * @throws CoreException
	 */
	private boolean saveDirtyEditors(IProject project, IWorkbench workbench) throws CoreException {
		// Note: we have to check for dirty editors in a loop
		// because Workbench.saveAllEditors() gives the user
		// the option of deselecting dirty files.  Doing this in
		// a loop is not elegant, but it does ensure that the
		// project's files really are all saved before continuing with
		// the submission.
		for (;;) {
			List dirtyEditors = getDirtyEditorsForProject(workbench, project);
			if (dirtyEditors.isEmpty()) {
				return true;
			}
			if (!workbench.saveAllEditors(true)) {
				// TODO: make it more obvious that the submit did not occur
				AutoCVSPlugin.getPlugin().getEventLog().logMessage(
						"Submission of project " + project.getName() + " cancelled");
				return false;
			}
		}
	}

	/**
	 * Get all dirty editors which are editing resources belonging to
	 * a project.
	 * @param workbench the workbench
	 * @param project the project
	 * @return the list of dirty editors for the project
	 */
	private List getDirtyEditorsForProject(IWorkbench workbench, IProject project) throws CoreException {
		Set projectResources = getProjectResources(project);
		List dirtyEditors = new LinkedList();
		
		// This code is based on Workbench.saveAllEditors().
		// You'd think there would be a much easier way
		// to find dirty editors.
		IWorkbenchWindow[] wwinList = workbench.getWorkbenchWindows();
		for (int i = 0; i < wwinList.length; i++) {
			IWorkbenchWindow wwin = wwinList[i];
			IWorkbenchPage[] pageList = wwin.getPages();
			for (int j = 0; j < pageList.length; j++) {
				IWorkbenchPage page = pageList[j];
				IEditorReference[] editorReferenceList = page.getEditorReferences(); 
				for (int k = 0; k < editorReferenceList.length; k++) {
					IEditorReference editorRef = editorReferenceList[k];
					IEditorPart editor = editorRef.getEditor(true);
					if (editor != null && editor.isDirty()) {
						IEditorInput input = editor.getEditorInput();
						IResource resource = (IResource) input.getAdapter(IResource.class);
						if (resource != null) {
							Debug.print("Got a resource from a dirty editor: " + resource.getName());
							if (projectResources.contains(resource))
								dirtyEditors.add(editor);
						}
					}
				}
			}
		}
		
		return dirtyEditors;
	}

	/**
	 * Get all resources in given project.
	 * @param project the project
	 * @return Set containing all resources in the project
	 */
	private Set getProjectResources(IProject project) throws CoreException {
		final Set projectResources = new HashSet();
		project.accept(new IResourceVisitor() {
			public boolean visit(IResource resource) throws CoreException {
				projectResources.add(resource);
				return true;
			}
		});
		return projectResources;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.IActionDelegate#selectionChanged(org.eclipse.jface.action.IAction, org.eclipse.jface.viewers.ISelection)
	 */
	public void selectionChanged(IAction action, ISelection selection) {
		this.selection = selection;
	}

}
